#  =============================================================================
#
#  @@-COPYRIGHT-START-@@
#
#  Copyright (c) 2022, Qualcomm Innovation Center, Inc. All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.
#
#  SPDX-License-Identifier: BSD-3-Clause
#
#  @@-COPYRIGHT-END-@@
#
#  =============================================================================
"""Pytorch Operation class and utilities"""

from typing import Union
import torch

import aimet_common.connected_graph.operation


class Op(aimet_common.connected_graph.operation.Op):
    """Subclass Op inherited from aimet_common.connected_graph.operation.Op"""

    def __init__(
        self,
        name: str,
        dotted_name: str,
        output_shape,
        is_anonymous: bool,
        op_type: str,
        residing_module: Union[torch.nn.Module, None],
    ):
        """
        Initializer for Op
        :param name: name of the operation
        :param dotted_name: dotted name of the operation
        :param output_shape: shape of the output product of the operation
        :param is_anonymous: whether this is an anonymous operation
        :param op_type: type of the operation
        :param residing_module: Torch module in which the op is situated
        """
        super().__init__(name, dotted_name, output_shape, is_anonymous, op_type)
        self.residing_module = residing_module

    def is_grid_preserving_op(self) -> bool:
        from .connectedgraph import ConnectedGraph
        from aimet_common.onnx._utils import _is_grid_preserving_op
        from ..onnx_utils import map_torch_types_to_onnx
        from ..nn import QuantizationMixin

        module = self.get_module()

        if not module:
            return self.type in ConnectedGraph.math_invariant_types

        if isinstance(module, QuantizationMixin):
            module = module.get_original_module()

        module_cls = type(module)
        onnx_op_types = map_torch_types_to_onnx.get(module_cls)

        if not onnx_op_types:
            # ONNX op type unknown.
            # To be safe, we should assume non-grid-preserving op in this case.
            return False

        return all(_is_grid_preserving_op(op_type) for op_type in onnx_op_types)
